from django.db.models import Count, Max
from django.conf import settings
from django.utils.translation import ugettext as _
from django.utils import timezone

from PIL import Image, ImageDraw, ImageFont

from models import Word, BingoField


# constants
COLOR_MODE_BLANK, COLOR_MODE_MARKED, COLOR_MODE_VOTED = range(3)

# settings
H_BOX_PADDING = getattr(settings, "HORIZONTAL_PADDING", 8)
V_BOX_PADDING = getattr(settings, "VERTICAL_PADDING", 30)
H_LINE_MARGIN = getattr(settings, "HORIZONTAL_LINE_MARGIN", 0)
V_LINE_MARGIN = getattr(settings, "VERTICAL_LINE_MARGIN", 5)
BORDER = getattr(settings, "BORDER", 1)

FONTPATH = getattr(settings, "FONT_PATH", "/path/to/font.ttf")
FONTSIZE = getattr(settings, "FONT_SIZE", 16)

NEUTRAL_FIELD_COLOR = getattr(settings, "NEUTRAL_FIELD_COLOR", (255, 255, 255))
NEUTRAL_WORD_COLOR = getattr(settings, "NEUTRAL_WORD_COLOR", (0, 0, 0))
MIDDLE_FIELD_COLOR = getattr(settings, "MIDDLE_FIELD_COLOR", (90, 90, 90))
MIDDLE_WORD_COLOR = getattr(settings, "MIDDLE_WORD_COLOR", (255, 255, 255))
# MARKED_FIELD_COLOR is stored in the BingoBoard model
MARKED_WORD_COLOR = getattr(settings, "MARKED_WORD_COLOR", (0, 0, 0))
# VOTED_FIELD_COLOR is stored in the BingoBoard model
VOTED_WORD_COLOR = getattr(settings, "VOTED_WORD_COLOR", (0, 0, 0))
VETO_FIELD_COLOR = getattr(settings, "VETO_FIELD_COLOR", (255, 255, 255))
VETO_WORD_COLOR = getattr(settings, "VETO_WORD_COLOR", (255, 0, 0))

BINGO_IMAGE_DATETIME_FORMAT = getattr(
    settings, "BINGO_IMAGE_DATETIME_FORMAT", "%Y-%m-%d %H:%M")

THUMBNAIL_WIDTH = getattr(settings, "THUMBNAIL_WIDTH", 300)
THUMBNAIL_HEIGHT = getattr(settings, "THUMBNAIL_HEIGHT", 80)


class Text(object):
    def __init__(self, draw, font, input):
        self.line_widths = []
        self.line_heights = []
        self.lines = input.split("\n")
        for line in self.lines:
            line_width, line_height = draw.textsize(line, font=font)
            self.line_widths.append(line_width + H_LINE_MARGIN)
            self.line_heights.append(line_height + V_LINE_MARGIN)

    def get_total_width(self):
        return sum(self.line_widths)

    def get_total_height(self):
        return sum(self.line_heights)


def get_texts(bingo_fields, font):
    im = Image.new("RGB", (200, 200), color=(255, 255, 255))
    draw = ImageDraw.Draw(im)
    words = []
    word_widths = []
    word_heights = []
    texts = []

    for bingo_field in bingo_fields:
        text = bingo_field.word.word
        if bingo_field.is_middle():
            text += _("\n{time}\nBingo #{board_id}").format(
                time=bingo_field.board.get_created(),
                board_id=bingo_field.board.board_id)
        texts.append(Text(draw, font, text))
    return texts


def get_colors(bingo_field, vote_counts, colormode=COLOR_MODE_BLANK):
    marked_field_color = bingo_field.board.color[1:]

    # convert from #rrggbb to decimal (r, g, b)
    marked_field_color = (
        int(marked_field_color[:2], 16),
        int(marked_field_color[2:4], 16),
        int(marked_field_color[4:6], 16),
    )
    # veto fields
    veto_field_color = (255, 255, 255)
    veto_word_color = (255, 0, 0)
    # border
    border_color = (0, 0, 0)

    field_color = NEUTRAL_FIELD_COLOR
    word_color = NEUTRAL_WORD_COLOR
    if bingo_field.is_middle():
        field_color = MIDDLE_FIELD_COLOR
        word_color = MIDDLE_WORD_COLOR
    elif colormode == COLOR_MODE_MARKED and bingo_field.vote:
        field_color = marked_field_color
        word_color = MARKED_WORD_COLOR
    elif colormode == COLOR_MODE_MARKED and bingo_field.vote is False:
        field_color = VETO_FIELD_COLOR
        word_color = VETO_WORD_COLOR
    elif colormode == COLOR_MODE_VOTED:
        max_votes = max(vote_counts.values())
        if max_votes > 0:
            votes = vote_counts[bingo_field.word.id]
            scaling = votes / float(max_votes)
            color_0 = int(255 - scaling * (255 - marked_field_color[0]))
            color_1 = int(255 - scaling * (255 - marked_field_color[1]))
            color_2 = int(255 - scaling * (255 - marked_field_color[2]))
            field_color = (color_0, color_1, color_2)
            word_color = VOTED_WORD_COLOR

    return field_color, word_color, border_color


def get_image(bingo_board, marked=False, voted=False):
    font = ImageFont.truetype(FONTPATH, FONTSIZE)

    all_bingo_fields = BingoField.objects.filter(
        board__game=bingo_board.game).select_related()
    bingo_fields = all_bingo_fields.filter(board=bingo_board).exclude(
        position=None).order_by("position")

    # image sizes
    texts = get_texts(bingo_fields, font)
    max_text_width = 0
    max_text_height = 0
    for text in texts:
        max_text_width = max(max_text_width, text.get_total_width())
        max_text_height = max(max_text_height, text.get_total_height())

    # inside the border
    inner_width = max_text_width + 2 * H_BOX_PADDING
    inner_height = max_text_height + 2 * V_BOX_PADDING

    # with border
    field_width = inner_width + 2 * BORDER
    field_height = inner_height + 2 * BORDER

    # total board width
    image_width = 5*(max_text_width + (2 * H_BOX_PADDING)) + (6 * BORDER)
    image_height = 5*(max_text_height + (2 * V_BOX_PADDING)) + (6 * BORDER)

    im = Image.new("RGB", (image_width, image_height), color=(255, 255, 255))
    draw = ImageDraw.Draw(im)

    # count votes per word, find the maximum
    vote_counts = {}
    for field in bingo_fields:
        vote_counts[field.word.id] = field.num_votes()

    # draw the board
    for bingo_field in bingo_fields:
        pos = bingo_field.position - 1
        y, x = divmod(pos, 5)
        # top left corner of the box
        box_left = x * (field_width - BORDER)
        box_top = y * (field_height - BORDER)
        # top left corner inside the box
        # inside the border, but outside of the padding
        word_left = box_left + BORDER
        word_top = box_top + BORDER

        # calculate the background and foreground color
        # for the field
        if marked:
            colormode = COLOR_MODE_MARKED
        elif voted:
            colormode = COLOR_MODE_VOTED
        else:
            colormode = COLOR_MODE_BLANK
        field_color, word_color, border_color = get_colors(
            bingo_field, vote_counts, colormode=colormode)

        # draw a border. if its more than 1px, its extended
        # by drawing smaller boxes inside
        for offset in xrange(BORDER):
            # -1 on the right/bottom corner is needed, because the line is
            # drawn below/right of the pixel, both for the
            # top/left and bottom/right corners
            draw.rectangle((
                (box_left + offset, box_top + offset),
                (box_left + field_width - offset - 1,
                    box_top + field_height - offset - 1)),
                fill=field_color,
                outline=border_color
            )

        v_center = (inner_height - texts[pos].get_total_height()) / 2.0
        text = texts[pos]
        h_offset = 0
        for idx, line in enumerate(text.lines):
            h_center = (inner_width - text.line_widths[idx]) / 2.0
            draw.text(
                (word_left + h_center,
                 word_top + v_center + h_offset),
                line, font=font, fill=word_color
            )
            h_offset += text.line_heights[idx]

    return im


def get_thumbnail(bingo_board, marked=False, voted=False,
                  thumbnail_width=THUMBNAIL_WIDTH,
                  thumbnail_height=THUMBNAIL_HEIGHT):
    im = get_image(bingo_board, marked, voted)
    im.thumbnail(
        size=(thumbnail_width, thumbnail_height), resample=Image.ANTIALIAS)
    return im



from django import forms
from django.utils.translation import pgettext, ugettext as _
from django.conf import settings
from django.contrib.auth.hashers import get_hasher

from models import BingoBoard, is_starttime
from times import is_starttime


SALT = getattr(settings, "SALT", "hackme")
GAME_START_TIMES = getattr(settings, "GAME_START_TIMES", None)

GAME_START_DISABLED = getattr(
    settings, "GAME_START_DISABLED", False)
GAME_DESCRIPTION_DISABLED = getattr(
    settings, "GAME_DESCRIPTION_DISABLED", False)


class CreateForm(forms.Form):
    password = forms.CharField(label=_(u"Password (optional)"),
                               widget=forms.PasswordInput(),
                               required=False)

    def __init__(self, *args, **kwargs):
        game = kwargs.get('game')
        # forms.Form throws an error on additional kwargs
        if 'game' in kwargs:
            del kwargs['game']
        super(CreateForm, self).__init__(*args, **kwargs)

        # add a game description field to the create form,
        # when there is no active game.
        if not GAME_DESCRIPTION_DISABLED and game is None:
            self.fields['description'] = forms.CharField(
                label=_(u'Game Description (optional)'),
                required=False)

    def clean_password(self):
        if self.cleaned_data['password']:
            hasher = get_hasher(algorithm='sha1')
            hashed_password = hasher.encode(
                self.cleaned_data['password'], SALT)
            return hashed_password
        else:
            return None

    def clean(self):
        if GAME_START_DISABLED:
            raise forms.ValidationError(
                _(u"Starting new games is disabled."))

        if not is_starttime():
            start, end = GAME_START_TIMES
            start_time_str = "{0}:{1}".format(
                str(start[0]).zfill(2),
                str(start[1]).zfill(2))
            end_time_str = "{0}:{1}".format(
                str(end[0]).zfill(2),
                str(end[1]).zfill(2))

            raise forms.ValidationError(
                _(u"Games can only be started between {0} and {1}.").format(
                    start_time_str, end_time_str))

        return super(CreateForm, self).clean()


class ReclaimForm(forms.Form):
    """
        Reclaim a board for the current game using the password.
        Works only for anonymous users.
    """
    password = forms.CharField(widget=forms.PasswordInput())

    def __init__(self, data=None, game=None, *args, **kwargs):
        self.game = game
        super(ReclaimForm, self).__init__(data=data, *args, **kwargs)

    def clean_password(self):
        hasher = get_hasher(algorithm='sha1')
        hashed_password = hasher.encode(self.cleaned_data['password'],
                                        SALT)
        if not self.game or self.game.is_expired():
            raise forms.ValidationError(
                _("The game is expired, please create a new board."))
        bingo_boards = BingoBoard.objects.filter(game=self.game,
                                                 password=hashed_password,
                                                 user=None)
        if bingo_boards.count() > 0:
            # if two users have the same password,
            # just return the first board.
            self.cleaned_data['bingo_board'] = bingo_boards[0]
        else:
            raise forms.ValidationError(
                _(u"No active board with this password."
                    " Try again, or create a new one."))
        return hashed_password


class ClaimForm(forms.Form):
    """
        Reclaim all old boards with the given password and assign them to
        a the user
    """
    password = forms.CharField(widget=forms.PasswordInput())

    def __init__(self, data=None, user=None, *args, **kwargs):
        self.user = user
        super(ClaimForm, self).__init__(data=data, *args, **kwargs)

    def clean_password(self):
        hasher = get_hasher(algorithm='sha1')
        hashed_password = hasher.encode(self.cleaned_data['password'],
                                        SALT)

        bingo_boards = BingoBoard.objects.filter(
            password=hashed_password).select_related()

        # check, that no User gets two BingoBoards from the same Game,
        # just because another User used the same password.
        # Games, which contain two BingoBoards with the given password
        # are filtered, so these BingoBoards cannot be claimed by anyone.
        game_ids = set()
        duplicate_game_ids = set()
        for board in bingo_boards:
            if board.game.id in game_ids:
                duplicate_game_ids.add(board.game.id)
            else:
                game_ids.add(board.game.id)

        # filter for user=None to prevent stealing already claimed boards
        bingo_boards.exclude(
            game__id__in=duplicate_game_ids).filter(
            user=None).update(user=self.user)

        return hashed_password


class ChangeThemeForm(forms.Form):
    theme = forms.CharField(required=False)


class RateGameForm(forms.Form):
    rating = forms.ChoiceField(
        choices=zip(["None"] + range(1, 6), [_("n/a")] + range(1, 6)))



from django.utils.translation import ugettext as _
from django.conf import settings
from django.contrib.staticfiles.storage import staticfiles_storage

static_url = staticfiles_storage.url


DEFAULT_THEMES = (
    ('Dark', 'bingo/themes/dark.css'),
)


def settings_context(request):
    """
        returns a dict with whitelisted settings
    """
    return {
    }


def themes(request):
    STATIC_ROOT = getattr(settings, "STATIC_ROOT", "")

    site_theme = getattr(settings, "THEME", "")
    if site_theme != '' and not site_theme.startswith("http://") \
            and not site_theme.startswith("https://"):
        site_theme = static_url(site_theme)

    themes = list(getattr(settings, "THEMES", DEFAULT_THEMES))
    if len(themes) > 0:
        themes = [(_("Default"), "")] + list(themes)
        for i in xrange(len(themes)):
            theme_name, theme_url = themes[i]
            if theme_url != '' and not theme_url.startswith("http://") \
                    and not theme_url.startswith("https://"):
                themes[i] = (theme_name, static_url(theme_url))
    else:
        # if there is no theme list, the user cannot
        # reset the theme, so we remove it from session
        # and use the default one.
        if 'theme' in request.session:
            del request.session['theme']

    user_theme = request.session.get('theme', '')

    return {
        'site_theme': site_theme,
        'themes': themes,
        'user_theme': user_theme,
    }


def bingo(request):
    sse_url = getattr(settings, "SSE_URL", None)
    use_sse = True if hasattr(settings, "USE_SSE") and sse_url else False
    polling_interval = getattr(settings, "POLLING_INTERVAL", 10)
    polling_interval_sse = getattr(settings, "POLLING_INTERVAL", 120)
    items = {
        'settings': settings_context(request),
        'use_sse': use_sse,
        'sse_url': sse_url,
        'polling_interval': polling_interval,
        'polling_interval_sse': polling_interval_sse,
    }
    for key, value in themes(request).items():
        items[key] = value

    return items



from django.shortcuts import render, redirect, get_object_or_404
from django.core.urlresolvers import reverse
from django.contrib.sites.shortcuts import get_current_site
from django.utils.translation import ugettext as _
from django.db import transaction
from django.http import HttpResponse, StreamingHttpResponse
from django.conf import settings
from django.core.cache import cache
from django.middleware.cache import CacheMiddleware
from django.contrib.auth.models import User
import json

from models import Word, Game, BingoBoard, BingoField, get_game
from forms import CreateForm, ClaimForm, ReclaimForm
from forms import ChangeThemeForm, RateGameForm
import image as image_module
import times

import logging
logger = logging.getLogger(__name__)

GAME_START_DISABLED = getattr(settings, "GAME_START_DISABLED", False)
THUMBNAIL_CACHE_EXPIRY = getattr(
    settings,
    "THUMBNAIL_CACHE_EXPIRY",
    5 * 60)
OLD_THUMBNAIL_CACHE_EXPIRY = getattr(
    settings,
    "OLD_THUMBNAIL_CACHE_EXPIRY",
    24 * 60 * 60)
USE_SSE = hasattr(settings, "SSE_URL")

if USE_SSE:
    REDIS_HOST = getattr(settings, "REDIS_HOST", None)
    REDIS_PORT = getattr(settings, "REDIS_PORT", None)
    try:
        from redis import Redis
        kwargs = {}
        if REDIS_HOST:
            kwargs['host'] = REDIS_HOST
        if REDIS_PORT:
            kwargs['port'] = REDIS_PORT
        redis = Redis(**kwargs)
    except ImportError:
        logger.warning(
            "USE_SSE is set, but no redis module found. Disabling SSE")
        USE_SSE = False


def _get_user_bingo_board(request):
    bingo_board = None
    game = get_game(site=get_current_site(request), create=False)
    session_board_id = request.session.get('board_id', None)
    user = request.user if request.user.is_authenticated() else None
    ip = request.META['REMOTE_ADDR']

    # try the board_id in the session
    if not session_board_id is None:
        try:
            bingo_board = BingoBoard.objects.get(id=session_board_id,
                                                 game=game)
            bingo_board.ip = ip
        except BingoBoard.DoesNotExist, e:
            pass
        # double board_id?
        except BingoBoard.MultipleObjectsReturned:
            pass

    if user:
        try:
            bingo_board = BingoBoard.objects.get(game=game, user=user)
        except BingoBoard.DoesNotExist, e:
            pass

    # no board_id in the session, try the ip
    if bingo_board is None:
        try:
            # user=None is required, to prevent anonymous users from
            # getting the board of a authenticated user on the same ip.
            bingo_board = BingoBoard.objects.get(game=game, ip=ip, user=None)
        except BingoBoard.DoesNotExist, e:
            pass

    if bingo_board is not None and not bingo_board.game.is_expired():
        BingoBoard.objects.filter(id=bingo_board.id).update(
            last_used=times.now())
        request.session['board_id'] = bingo_board.id

    return bingo_board


def _get_image_name(board_id, marked=False, voted=False):
    if voted:
        filename = _("board_{0}_voted").format(board_id)
    elif marked:
        filename = _("board_{0}_marked").format(board_id)
    else:
        filename = _("board_{0}").format(board_id)

    return filename


def _publish_num_users(site_id, num_users=None, num_active_users=None):
    if num_users is not None:
        redis.publish("num_users", json.dumps({
            'site_id': site_id,
            'num_users': num_users,
        }))
    if num_active_users is not None:
        redis.publish("num_active_users", json.dumps({
            'site_id': site_id,
            'num_active_users': num_active_users,
        }))


def main(request, reclaim_form=None, create_form=None):
    game = get_game(site=get_current_site(request), create=False)
    bingo_board = _get_user_bingo_board(request)

    create_form = CreateForm(prefix="create", game=game)
    reclaim_form = ReclaimForm(prefix="reclaim")

    boards = BingoBoard.objects.filter(game=game)
    old_games = Game.objects.filter(
        site=get_current_site(request)
    ).exclude(
        # games without any boards created
        bingoboard=None
    ).exclude(
        # games without game_id are hidden
        game_id=None
    ).order_by("-created")

    if game is not None:
        old_games = old_games.exclude(id=game.id)

    return render(request, "bingo/main.html", {
        'my_board': bingo_board,
        'create_form': create_form,
        'reclaim_form': reclaim_form,
        'boards': boards,
        'current_game': game,
        'old_games': old_games,
        'can_start_game': not GAME_START_DISABLED,
        })


def game(request, game_id):
    bingo_board = _get_user_bingo_board(request)
    return render(request, "bingo/game.html", {
        'game': get_object_or_404(
            Game, site=get_current_site(request), game_id=game_id),
        'my_board': bingo_board
    })


def profile(request, username):
    user = get_object_or_404(User, username=username)
    form = ClaimForm(request.POST or None, user=user)
    form.is_valid()
    boards = user.bingoboard_set.filter(game__site=get_current_site(request))
    return render(request, "bingo/profile.html",
            {'profile_user': user, 'boards': boards, 'claim_form': form})


def reclaim_board(request):
    ip = request.META['REMOTE_ADDR']
    game = get_game(site=get_current_site(request), create=False)
    if game is not None:
        Game.objects.filter(id=game.id).update(last_used=times.now())
    bingo_board = _get_user_bingo_board(request)

    if not bingo_board is None:
        return redirect(reverse(bingo, kwargs={
            'board_id': bingo_board.board_id}))
    if request.POST:
        reclaim_form = ReclaimForm(request.POST, game=game, prefix="reclaim")
        if reclaim_form.is_valid():
            bingo_board = reclaim_form.cleaned_data['bingo_board']
            request.session['board_id'] = bingo_board.id
            bingo_board.ip = ip
            bingo_board.save()
            return redirect(reverse(bingo, kwargs={'board_id':
                                                   bingo_board.board_id}))
    else:
        reclaim_form = ReclaimForm(prefix="reclaim")
    create_form = CreateForm(prefix="create", game=game)
    return render(request,
                  "bingo/reclaim_board.html", {
                      'reclaim_form': reclaim_form,
                      'create_form': create_form,
                  })


def create_board(request):
    bingo_board = _get_user_bingo_board(request)
    game = bingo_board.game if bingo_board is not None else None
    user = request.user

    if bingo_board:
        Game.objects.filter(id=bingo_board.game.id).update(
            last_used=times.now())
        return redirect(reverse(bingo, kwargs={
            'board_id': bingo_board.board_id}))
    elif request.POST:
        create_form = CreateForm(
            request.POST,
            prefix="create",
            game=game)

        if create_form.is_valid():
            with transaction.atomic():
                ip = request.META['REMOTE_ADDR']
                password = create_form.cleaned_data.get('password')
                game_description = create_form.cleaned_data.get(
                    'description', '')
                game = get_game(
                    site=get_current_site(request),
                    description=game_description,
                    create=True)
                Game.objects.filter(id=game.id).update(
                    last_used=times.now())

                # if the user is logged in, associate the board with
                # the user, and ignore the password
                # (so no other user can claim the board)
                user = user if user.is_authenticated() else None
                if user:
                    password = None

                bingo_board = BingoBoard(
                    game=game, user=user, ip=ip, password=password)
                bingo_board.save()

                if USE_SSE:
                    _publish_num_users(game.site.id, game.num_users(),
                                       game.num_active_users())

                return redirect(reverse(bingo, kwargs={
                    'board_id': bingo_board.board_id}))
        else:
            reclaim_form = ReclaimForm(prefix="reclaim")
            return render(
                request,
                "bingo/reclaim_board.html",
                {
                    'reclaim_form': reclaim_form,
                    'create_form': create_form,
                }
            )
    else:
        return redirect(reverse(main))


def bingo(request, board_id=None):
    bingo_board = get_object_or_404(
        BingoBoard, board_id=board_id,
        game__site=get_current_site(request)
    )
    my_bingo_board = _get_user_bingo_board(request)
    fields_on_board = bingo_board.get_board_fields().select_related()
    all_word_fields = bingo_board.get_all_word_fields().order_by(
        "word__word").select_related()

    return render(request, "bingo/bingo.html", {
        "fields_on_board": fields_on_board,
        "board": bingo_board,
        "my_board": my_bingo_board,
        "all_word_fields": all_word_fields,
        "rate_form": RateGameForm(),
        })


class VoteException(Exception):
    """
        something went wrong voting on a field.
        (i.e. the field does not belong to the user's board)
    """
    pass


def _post_vote(user_bingo_board, field, vote):
    """
        change vote on a field
        @param user_bingo_board: the user's bingo board or None
        @param field: the BingoField to vote on
        @param vote: the vote property from the HTTP POST
        @raises: VoteException: if user_bingo_board is None or
        the field does not belong to the user's bingo board.
        @raises Http404: If the BingoField does not exist.
    """

    if field.board != user_bingo_board:
        raise VoteException(
            "the voted field does not belong to the user's BingoBoard")

    if vote == "0":
        field.vote = None
    elif vote == "+":
        field.vote = True
    elif vote == "-":
        field.vote = False
    field.save()

    # update last_used with current timestamp
    game = user_bingo_board.game
    Game.objects.filter(id=game.id).update(
        last_used=times.now())

    # invalidate vote cache
    vote_counts_cachename = 'vote_counts_game={0:d}'.format(
        field.board.game.id)
    cache.delete(vote_counts_cachename)

    # publish the new vote counts for server-sent events
    if USE_SSE:
        votes = field.num_votes()
        redis.publish("word_votes", json.dumps({
            'site_id': game.site.id,
            'word_id': field.word.id,
            'vote_count': votes,
        }))
        redis.publish("field_vote", json.dumps({
            'site_id': game.site.id,
            'field_id': field.id,
            'vote': vote,
        }))


def vote(request, ajax, board_id=None):
    user_bingo_board = _get_user_bingo_board(request)
    field = None

    # post request: update field.vote
    if "field_id" in request.POST and times.is_after_votetime_start():
        field_id = request.POST.get("field_id", 0)
        field = get_object_or_404(BingoField, id=field_id)
        vote = request.POST.get("vote", "0")
        try:
            _post_vote(user_bingo_board, field, vote)
        except VoteException:
            pass  # ignore the vote

    # form submitted without ajax, redirect back to the bingo page
    if not ajax:
        if field:
            return redirect(
                reverse(bingo, kwargs={"board_id": field.board.board_id}))
        else:
            return redirect(reverse(main))

    # page submitted with ajax, return vote data as json
    if board_id is not None:  # board id given in the url
        bingo_board = get_object_or_404(BingoBoard, id=board_id)
    elif user_bingo_board is not None:  # user's bingo board
        bingo_board = user_bingo_board
    else:  # missing the needed parameters
        # Set data to {} to prevent an AttributeError
        return HttpResponse(json.dumps({}), content_type="application/json")

    # add data about the game
    game = bingo_board.game
    num_users = game.num_users()
    num_active_users = game.num_active_users()
    data = {
        'num_users': num_users,
        'num_active_users': num_active_users,
        'is_expired': bingo_board.game.is_expired(),
    }

    # send user number, if changed
    if USE_SSE:
        old_num_users = cache.get(
            "num_users__site={0:d}".format(game.site.id))
        old_num_active_users = cache.get(
            "num_active_users__site={0:d}".format(game.site.id))

        if old_num_users != num_users:
            cache.set("num_users__site={0:d}".format(game.site.id),
                num_users)
            _publish_num_users(game.site.id, num_users=num_users)
        if old_num_active_users != num_active_users:
            cache.set("num_active_users__site={0:d}".format(game.site.id),
                num_active_users)
            _publish_num_users(game.site.id, num_active_users=num_active_users)

    for field in bingo_board.bingofield_set.all():
        # None="0", "+"=vote, "-"=veto
        vote = "0" if field.vote is None else "+" if field.vote else "-"
        data[field.id] = (vote, field.num_votes())

    return HttpResponse(json.dumps(data), content_type="application/json")


def rate_game(request):
    bingo_board = _get_user_bingo_board(request)
    if bingo_board:
        form = RateGameForm(request.POST)
        if form.is_valid():
            rating = form.cleaned_data['rating']
            if rating == "None":
                rating = None
            BingoBoard.objects.filter(id=bingo_board.id).update(rating=rating)
        return redirect(reverse(bingo, kwargs={
            'board_id': bingo_board.board_id}))
    else:
        return redirect(reverse(main))


def image(request, board_id, marked=False, voted=False):
    bingo_board = get_object_or_404(
        BingoBoard, board_id=board_id,
        game__site=get_current_site(request))
    response = HttpResponse(content_type="image/png")
    filename = _get_image_name(board_id, marked, voted) + ".png"
    response['Content-Disposition'] = 'filename={0}'.format(filename)
    im = image_module.get_image(bingo_board, marked, voted)
    im.save(response, "png")
    return response


def thumbnail(request, board_id, marked=False, voted=False):
    bingo_board = get_object_or_404(
        BingoBoard, board_id=board_id,
        game__site=get_current_site(request))

    # check if the board is from an expired game
    game_expired_cachename = "game_expired__board={0:d}".format(
        int(bingo_board.id))
    game_expired = cache.get(
        game_expired_cachename)
    if game_expired is None:
        game_expired = bingo_board.game.is_expired()
        cache.set(
            game_expired_cachename,
            game_expired, 60 * 60)

    # when the game of the board is expired,
    # the thumbnail can be cached longer.
    if game_expired:
        m = CacheMiddleware(cache_timeout=OLD_THUMBNAIL_CACHE_EXPIRY,
            key_prefix="site_id={0:d}".format(get_current_site(request).id))
    else:
        m = CacheMiddleware(cache_timeout=THUMBNAIL_CACHE_EXPIRY,
            key_prefix="site_id={0:d}".format(get_current_site(request).id))

    response = m.process_request(request)
    if response:  # cached
        return response

    response = HttpResponse(content_type="image/png")
    filename = _get_image_name(board_id, marked, voted) + \
        "_" + _("thumbnail") + ".png"
    response['Content-Disposition'] = 'filename={0}'.format(filename)
    im = image_module.get_thumbnail(bingo_board, marked, voted)
    im.save(response, "png")

    return m.process_response(request, response)


def change_theme(request):
    if request.POST:
        form = ChangeThemeForm(request.POST)
        if form.is_valid():
            request.session['theme'] = form.cleaned_data['theme']
    return redirect(reverse('bingo.views.main'))



#!/usr/bin/env python
import json
import sys
import os

from sse import Sse
from flask import Flask, Response
from redis import Redis
from gevent import Timeout

redis = Redis()


SITE_ID = int(os.environ['SITE_ID']) if 'SITE_ID' in os.environ else 1
TIMEOUT = 120


application = Flask(__name__)
event_types = ['word_votes', 'field_vote', 'num_users', 'num_active_users']


def eventstream():
    sse = Sse()
    while True:
        pubsub = redis.pubsub()
        for event in event_types:
            pubsub.subscribe(event)
        try:
            with Timeout(TIMEOUT) as timeout:
                for message in pubsub.listen():
                    if message['type'] != "message":
                        continue
                    data = json.loads(message['data'])
                    if not 'site_id' in data or data['site_id'] != SITE_ID:
                        continue

                    sse.add_message(message['channel'], str(message['data']))
                    for event in sse:
                        yield str(event)
                    sse.flush()

                    timeout.cancel()
                    timeout.start()

        # heartbeat, to detect if a user is disconnected
        except Timeout, t:
            if t is not timeout:  # not our timeout
                raise
            yield ":\n\n"  # heartbeat message


@application.route("/")
def events():
    response = Response(eventstream(), mimetype="text/event-stream")
    response.headers.add("Access-Control-Allow-Origin", "*")
    return response


if __name__ == "__main__":
    application.run(threaded=True)



from django.utils import timezone
from django.conf import settings

import pytz
from datetime import datetime

# Time range, in which a game can be started. None = no limit
# or a ((Hour, Minute), (Hour, Minute)) tuple defining the range.
GAME_START_TIMES = getattr(settings, "GAME_START_TIMES", None)

# Time, after which a running game is ended. Has only effect, if
# GAME_START_TIMES is set, and needs to be outside of GAME_START_TIMES.
# Values: tuple (hour, minute) or None for no restriction
GAME_END_TIME = getattr(settings, "GAME_END_TIME", None)

# Time, after which voting is possible.
# Needs to come after the start time start.
VOTE_START_TIME = getattr(settings, "VOTE_START_TIME", None)


def now():
    return timezone.get_current_timezone().normalize(timezone.now())


def get_times():
    time_now = now()

    start_time_start = None
    start_time_end = None
    end_time = None
    vote_start_time = None
    if GAME_START_TIMES:
        start, end = GAME_START_TIMES
        start_time_start = time_now.replace(
            hour=start[0], minute=start[1], second=0, microsecond=0)
        start_time_end = time_now.replace(
            hour=end[0], minute=end[1], second=0, microsecond=0)

        # when the end of start time is "before" the start of start time,
        # the end of start time is tomorrow
        if start_time_end < start_time_start:
            start_time_end += timezone.timedelta(1, 0)

    if GAME_END_TIME:
        end = GAME_END_TIME
        end_time = time_now.replace(
            hour=end[0], minute=end[1], second=0, microsecond=0)

        # when the end time is "before" the end of starttime_end,
        # the game ends tomorrow
        if start_time_end is not None and end_time < start_time_end:
            end_time = end_time + timezone.timedelta(1, 0)

    if VOTE_START_TIME:
        vote_start = VOTE_START_TIME
        vote_start_time = time_now.replace(
            hour=vote_start[0], minute=vote_start[1], second=0, microsecond=0)

        # The vote time must come after the start of starttime.
        # If it comes before, it must be tomorrow
        if start_time_start is not None and vote_start_time < start_time_start:
            vote_start_time = vote_start_time + timezone.timedelta(1, 0)

    # some sanity checks
    if GAME_START_TIMES and VOTE_START_TIME:
        start_time_start < vote_start_time
    if GAME_END_TIME and VOTE_START_TIME:
        assert end_time > vote_start_time
    if GAME_START_TIMES and GAME_END_TIME:
        assert end_time > start_time_end

    return {
        'now': time_now,
        'start_time_start': start_time_start,
        'start_time_end': start_time_end,
        'end_time': end_time,
        'vote_start_time': vote_start_time,
    }


def get_endtime():
    """ returns the (static) game end time """
    return get_times()['end_time']


def is_starttime():
    """
        returns True, if no start times are set, or the current time
        lies inside the starttime.
    """
    if GAME_START_TIMES is None:
        return True
    else:
        times = get_times()
        return times['start_time_start'] \
            < times['now'] \
            < times['start_time_end']


def is_after_votetime_start():
    """
        returns True, if no VOTE_START_TIME is set,
        or the current time is after the start of vote
        time
    """
    if VOTE_START_TIME is None:
        return True
    else:
        times = get_times()
        return times['vote_start_time'] <= times['now']


def is_after_endtime():
    if GAME_END_TIME is None or is_starttime():
        return False
    else:
        times = get_times()
        return times['end_time'] < times['now']



from django.contrib.auth.models import User
from django.utils.translation import ugettext as _
from django.core.exceptions import ValidationError
from django.conf import settings
from django.utils import timezone
from django.db import models, transaction
from django.contrib.sites.models import Site
from django.core.cache import cache
from django.core.urlresolvers import reverse

from colorful.fields import RGBColorField

from random import randint

from times import is_starttime, is_after_endtime, get_endtime


# Color ranges
COLOR_FROM = getattr(settings, "COLOR_FROM", 80)
COLOR_TO = getattr(settings, "COLOR_TO", 160)

# Expire game after hard-timeout seconds, or
# soft-timeout seconds inactivity, whatever comes first.
GAME_SOFT_TIMEOUT = getattr(settings, "GAME_SOFT_TIMEOUT", 60)
GAME_HARD_TIMEOUT = getattr(settings, "GAME_HARD_TIMEOUT", 120)

BINGO_IMAGE_DATETIME_FORMAT = getattr(
    settings, "BINGO_IMAGE_DATETIME_FORMAT", "%Y-%m-%d %H:%M")

# if a user did not vote/refresh his board for this time,
# he is counted as inactive
USER_ACTIVE_TIMEOUT = getattr(settings, "USER_ACTIVE_TIMEOUT", 5)

THUMBNAILS_ENABLED = getattr(settings, "THUMBNAILS_ENABLED", True)


class Word(models.Model):
    """
        a word entry. Should not be deleted, but only disabled
        to preserve old BingoFields
    """
    word = models.CharField(max_length=255, unique=True)
    description = models.CharField(max_length=255, blank=True)
    is_middle = models.BooleanField(default=False)
    site = models.ManyToManyField(Site, blank=True)

    class Meta:
        ordering = ("word",)

    def __unicode__(self):
        return u"Word: " + self.word


class TimeRangeError(Exception):
    pass


def get_game(site, description="", create=False):
    """
        get the current game, if its still active, else
        creates a new game, if the current time is inside the
        GAME_START_TIMES interval and create=True
        @param create: create a game, if there is no active game
        @returns: None if there is no active Game, and none shoul be
        created or the (new) active Game.
    """

    game = None
    games = Game.objects.filter(site=site).order_by("-created")
    try:
        game = games[0]
    except IndexError:
        game = None

    # no game, yet, or game expired
    if game is None or game.is_expired() or is_after_endtime():
        if create:
            if is_starttime():
                game = Game(site=site, description=description)
                game.save()
            else:
                raise TimeRangeError(
                    _(u"game start outside of the valid timerange"))
        else:
            game = None

    # game exists and its not after the GAME_END_TIME
    elif not is_after_endtime():
        game = games[0]

    return game


class Game(models.Model):
    game_id = models.IntegerField(blank=True, null=True)
    description = models.CharField(max_length=255, blank=True)
    site = models.ForeignKey(Site)
    created = models.DateTimeField(auto_now_add=True)
    last_used = models.DateTimeField(auto_now_add=True)

    class Meta:
        unique_together = ("game_id", "site")

    def get_absolute_url(self):
        return reverse('bingo.views.game', kwargs={"game_id": self.game_id})

    def __unicode__(self):
        return _(u"Game #{0} created at {1} (site {2})").format(
            self.game_id,
            timezone.localtime(self.created).strftime(u"%Y-%m-%d %H:%M"),
            self.site)

    def hard_expiry(self):
        if GAME_HARD_TIMEOUT is not None:
            return self.created + timezone.timedelta(0, GAME_HARD_TIMEOUT * 60)

    def soft_expiry(self):
        if GAME_SOFT_TIMEOUT is not None:
            return self.last_used + timezone.timedelta(
                0, GAME_SOFT_TIMEOUT * 60)

    def end_time(self):
        hard_expiry = self.hard_expiry()
        end_time = get_endtime()
        if hard_expiry is not None and end_time is not None:
            return min(end_time, hard_expiry)
        elif hard_expiry is None:
            # may be None
            return end_time
        else:
            # may be None
            return hard_expiry

    def is_expired(self):
        # game expired, because no one used it
        now = timezone.now()
        diff_last_used = now - self.last_used
        diff_created = now - self.created

        seconds_since_last_used = \
            diff_last_used.days * 24 * 60 * 60 + diff_last_used.seconds
        seconds_since_created = \
            diff_created.days * 24 * 60 * 60 + diff_created.seconds

        # game expired, because nobody used it
        if GAME_SOFT_TIMEOUT and seconds_since_last_used \
                > (GAME_SOFT_TIMEOUT * 60):
            return True

        # game expired, because its too old, even when someone is using it
        elif GAME_HARD_TIMEOUT and seconds_since_created \
                > (GAME_HARD_TIMEOUT * 60):
            return True
        else:
            return False

    def num_users(self):
        return self.bingoboard_set.count()

    def num_active_users(self):
        if USER_ACTIVE_TIMEOUT:
            return self.bingoboard_set.exclude(
                last_used__lte=timezone.now()
                - timezone.timedelta(0, 60 * USER_ACTIVE_TIMEOUT)
            ).count()
        else:
            return self.num_users()

    def num_votes_word(self, word_id):
        """
            get the (up, down) votes for a word

            @param word_id: the id of the word
            @returns (up, down) tuple with vote counts
        """

        # try to get it from cache
        vote_counts_cachename = \
            'vote_counts_game={0:d}'.format(
                self.id)
        vote_counts = cache.get(vote_counts_cachename)

        # vote counts not in cache or word not in the dict
        if vote_counts is None or word_id not in vote_counts:

            # querys for number of up / down votes
            vote_counts_up = Word.objects.filter(
                bingofield__vote=True, bingofield__board__game=self)\
                .annotate(num=models.Count('bingofield'))
            vote_counts_down = Word.objects.filter(
                bingofield__vote=False, bingofield__board__game=self)\
                .annotate(num=models.Count('bingofield'))

            # fetch all at once, so its only one query for up, one for down
            vote_counts_up = dict(
                [(word.id, word.num) for word in vote_counts_up])
            vote_counts_down = dict(
                [(word.id, word.num) for word in vote_counts_down])

            vote_counts = {}
            # words with no up or down vote
            for word in Word.objects.filter(bingofield__board__game=self):
                if not word.id in vote_counts_up:
                    vote_counts_up[word.id] = 0
                if not word.id in vote_counts_down:
                    vote_counts_down[word.id] = 0
                vote_counts[word.id] = {
                    'up': vote_counts_up[word.id],
                    'down': vote_counts_down[word.id],
                }

            cache.set(vote_counts_cachename, vote_counts)

        return (vote_counts[word_id]['up'], vote_counts[word_id]['down'])

    def rating(self):
        return self.bingoboard_set.exclude(rating=None).aggregate(
            rating=models.Avg('rating'))['rating'] or 0

    def num_ratings(self):
        return self.bingoboard_set.exclude(rating=None).aggregate(
            num=models.Count('rating'))['num']

    def save(self):
        if self.pk is None:
            games = Game.objects.filter(site=self.site)
            current_id = games.aggregate(
                max_id=models.Max('game_id'))['max_id']
            if current_id is None:
                self.game_id = 1
            else:
                self.game_id = current_id + 1
        return super(Game, self).save()


def _get_random_words(site):
    all_words = Word.objects.filter(site=site).order_by("?")
    middle_words = all_words.filter(site=site, is_middle=True).order_by("?")
    words = all_words.filter(is_middle=False)
    if middle_words.count() == 0:
        raise ValidationError(_(u"No middle words in database"))
    middle = middle_words[0]
    if words.count() < 24:
        raise ValidationError(_(u"Not enough (non-middle) words in database"))
    return list(words), middle


RATINGS = (
    (None, _("unrated")),
    (1, _("{0} stars").format(1)),
    (2, _("{0} stars").format(2)),
    (3, _("{0} stars").format(3)),
    (4, _("{0} stars").format(4)),
    (5, _("{0} stars").format(5))
)


class BingoBoard(models.Model):
    board_id = models.IntegerField(blank=True, null=True)
    game = models.ForeignKey("Game")
    color = RGBColorField()
    ip = models.GenericIPAddressField(blank=True, null=True)
    user = models.ForeignKey(User, blank=True, null=True)
    password = models.CharField(max_length=255, blank=True, null=True)
    created = models.DateTimeField(auto_now_add=True)
    last_used = models.DateTimeField(auto_now_add=True)
    rating = models.IntegerField(choices=RATINGS, null=True, blank=True,
        default=None)

    class Meta:
        ordering = ("-board_id",)
        # board_id, game__site is not possible
        unique_together = ("board_id", "game")

    def get_absolute_url(self):
        return reverse('bingo.views.bingo', kwargs={"board_id": self.board_id})

    def clean(self):
        try:
            board = BingoBoard.objects.get(
                board_id=self.board_id, game__site=self.game.site)
            # if its not an edit of the same BingoBoard
            if board.pk != self.pk:
                raise ValidationError(
                    _(u"BingoBoard ID not unique for site {0}").format(
                        self.game.site))
        except BingoBoard.DoesNotExist:
            return  # everything okay

    def save(self):
        if self.ip is None and self.user is None:
            raise ValidationError(
                _(u"BingoBoard must have either an ip or an user"))

        if self.pk is None:
            # unique_together for optional fields
            game_boards = BingoBoard.objects.filter(game=self.game)
            if not self.user is None:
                if game_boards.filter(user=self.user).count() > 0:
                    raise ValidationError(
                        _(u"game and user must be unique_together"))
            # ip only needs to be unique for anonymous users
            # to check if the ip is relevant:
            # - check the current board does not have an user
            # - filter for existing boards from this ip only with user=None
            if not self.ip is None and not self.user:
                if game_boards.filter(ip=self.ip, user=None).count() > 0:
                    raise ValidationError(
                        _(u"game and ip must be unique_together"))

            # generate a color
            self.color = "#%x%x%x" % (
                randint(COLOR_FROM, COLOR_TO),
                randint(COLOR_FROM, COLOR_TO),
                randint(COLOR_FROM, COLOR_TO))

            # first create the fields, so the board will
            # not be saved, when field creation fails
            fields = self.create_bingofields()

            # then create a board_id
            with transaction.atomic():
                bingo_boards = BingoBoard.objects.filter(
                    game__site=self.game.site)
                current_id = bingo_boards.aggregate(
                    max_id=models.Max('board_id'))['max_id']
                if current_id is None:
                    self.board_id = 1
                else:
                    self.board_id = current_id + 1

                # create the board
                super(BingoBoard, self).save()

                # now that the board has a pk, save the fields
                for field in fields:
                    field.board = self
                    field.save()
        else:
            super(BingoBoard, self).save()

    def create_bingofields(self):
        count = 0
        words, middle = _get_random_words(site=self.game.site)
        fields = []
        for i in xrange(25):
            # 13th field = middle
            if i == 12:
                fields.append(BingoField(word=middle, position=i+1))
            else:
                fields.append(BingoField(word=words[count], position=i+1))
                count += 1
        # create fields without position for every
        # active word, which is not on the board, too.
        # the index is 24, because there are only 24
        # non-middle words on the field
        assert count == 24
        for word in words[24:]:
            fields.append(BingoField(word=word, position=None))

        return fields

    def get_board_fields(self):
        return self.bingofield_set.exclude(position=None).order_by("position")

    def get_all_word_fields(self):
        return self.bingofield_set.filter(word__is_middle=False)

    def get_created(self):
        """
            get created field with BINGO_IMAGE_DATETIME_FORMAT formatting
        """
        return timezone.localtime(self.created).strftime(
            BINGO_IMAGE_DATETIME_FORMAT)

    def get_last_used(self):
        """
            get last_used field with BINGO_IMAGE_DATETIME_FORMAT formatting
        """
        return timezone.localtime(self.last_used).strftime(
            BINGO_IMAGE_DATETIME_FORMAT)

    def num_votes(self, field):
        """
            get the up/down votes for a given field of THIS BingoBoard
            @param field: a BingoField object
            @returns (up, down)
        """

        # map BingoField to Word, for one BingoBoard
        # so the cached vote counts for Word can be used
        field2word_cachename = "field2word_board={0:d}".format(self.id)
        field2word = cache.get(field2word_cachename) or {}

        # create, if not cached
        if field2word == {} or field.id not in field2word:
            for loop_field in self.bingofield_set.all().select_related():
                field2word[loop_field.id] = loop_field.word.id
            cache.set(field2word_cachename, field2word, 60*60)

        # get the Word.id for field
        word_id = field2word[field.id]
        return self.game.num_votes_word(word_id)

    def thumbnails_enabled(self):
        return THUMBNAILS_ENABLED

    def __unicode__(self):
        return _(u"BingoBoard #{0} created by {1} (site {2})").format(
            self.board_id,
            self.user if self.user else self.ip,
            self.game.site)


def position_validator(value):
    if not value is None and not (0 < value < 26):
        raise ValidationError(_(
            _(u"invalid position. valid values range: 1-25 or None")))


class BingoField(models.Model):
    word = models.ForeignKey("Word")
    board = models.ForeignKey("BingoBoard")
    position = models.SmallIntegerField(
        validators=[position_validator],
        blank=True, null=True, default=None)
    vote = models.NullBooleanField(default=None)

    class Meta:
        ordering = ("board", "-position")

    def is_middle(self):
        return self.position == 13

    def num_votes(self):
        positive, negative = self.board.num_votes(self)
        return max(0, positive - negative)

    def clean(self):
        if self.is_middle() and not self.word.is_middle:
            raise ValidationError(_(
                u"The BingoField has middle position, "
                u"but the word is no middle word"))
        elif not self.is_middle() and self.word.is_middle:
            raise ValidationError(_(
                u"The BingoField is not in the middle, "
                u"but the word is a middle word"))

    def __unicode__(self):
        if self.position is not None:
            return _(u"BingoField: word={0}, pos=({1},{2}){3})").format(
                self.word, self.position/5+1, self.position % 5,
                _(u" [middle]") if self.is_middle() else u"")
        else:
            return _(u"BingoField: word={0}, (not on the board)").format(
                self.word)



from django.views.generic import ListView
from django.contrib.sites.models import Site

from models import Game


class GameList(ListView):
    queryset = Game.objects.filter(
        site=Site.objects.get_current()).exclude(
        bingoboard=None).order_by("-created")
    paginate_by = 25



from django.conf.urls import patterns, include, url

import views
import generic

urlpatterns = patterns(
    '',
    url('^$', views.main),
    url('^bingo/reclaim/$', views.reclaim_board),
    url('^bingo/create/$', views.create_board),
    url('^bingo/vote/$', views.vote, {"ajax": False}),
    url('^bingo/rate/$', views.rate_game),
    url('^ajax/vote/$', views.vote, {"ajax": True}),
    url('^ajax/vote/(?P<board_id>[0-9]+)/$', views.vote, {"ajax": True}),
    url('^bingo/(?P<board_id>[0-9]+)/$', views.bingo),
    url('^game/(?P<game_id>[0-9]+)/$', views.game),
    url('^game/$', generic.GameList.as_view(), name="games_list"),
    url('^thumbnail/(?P<board_id>[0-9]+)/marked/$',
        views.thumbnail, {"marked": True}),
    url('^thumbnail/(?P<board_id>[0-9]+)/voted/$',
        views.thumbnail, {"voted": True}),
    url('^thumbnail/(?P<board_id>[0-9]+)/$', views.thumbnail),
    url('^image/(?P<board_id>[0-9]+)/marked/$', views.image, {"marked": True}),
    url('^image/(?P<board_id>[0-9]+)/voted/$', views.image, {"voted": True}),
    url('^image/(?P<board_id>[0-9]+)/$', views.image),
    url('^users/(?P<username>[^/]+)/$', views.profile),
    url('^change_theme/$', views.change_theme),
)



#!/usr/bin/env python
import os
import sys
from setuptools import setup
from setuptools.command.install_lib import install_lib as _install_lib


with open('requirements.txt') as f:
    required = f.read().splitlines()


class install_lib(_install_lib):
    def run(self):
        from django.core.management.commands.compilemessages \
            import compile_messages
        os.chdir('bingo')
        compile_messages(sys.stderr)
        os.chdir("..")

setup(name='django-bingo',
      description='Bingo',
      long_description='Bingo game implemented in django',
      author='Alexander Schier',
      author_email='allo@laxu.de',
      url="https://github.com/allo-/django-bingo",
      version='1.6.0',
      packages=['bingo'],
      package_data={'bingo': ['templates/bingo/*.*',
                              'static/bingo/*.*',
                              'static/bingo/js/*.*',
                              'locale/*/LC_MESSAGES/*.*']},
      scripts=['bingo/bin/django-bingo-serversent.py'],
      include_package_data=True,
      install_requires=required,
      classifiers=[
          'Framework :: Django',
          'Topic :: Games/Entertainment :: Board Games',
          'License :: OSI Approved :: GNU Affero General Public License v3',
          'Operating System :: OS Independent',
          'Programming Language :: Python'
      ]
      )



from django.contrib import admin
from django.db import models
from django import forms

from models import *


def bingoboard_user(bingo_board):
    return bingo_board.user if bingo_board.user else bingo_board.ip


def bingoboard_name(bingo_board):
    return u'BingoBoard #{0} (site {1})'.format(
        bingo_board.board_id, bingo_board.game.site)


def word_sites(word):
    """
        @returns: the sites, where a Word instance is active,
        truncated, so at most 5 sites are displayed
    """
    sites = map(lambda site: site.domain, word.site.all()[:5])
    return ", ".join(sites if len(sites) <= 5 else sites+["[...]"])


def game_id(game):
    """
        just the game.id prefixed by "Game #"
        so its easier to click in the admin interface
    """
    return u"Game #{0}".format(game.game_id)


def bingoboard_game_id(bingo_board):
    """
        @returns the game.id for the game of a BingoBoard instance
    """
    return game_id(bingo_board.game)


bingoboard_name.short_description = "Bingo"
bingoboard_game_id.short_description = "Game"
word_sites.short_description = "Sites"


class WordAdmin(admin.ModelAdmin):
    list_filter = ("site", "site", "is_middle")
    list_display = ("word", word_sites, "is_middle")
    search_fields = ("word", )

    class Meta:
        ordering = ("word",)

    def has_delete_permission(self, request, obj=None):
        return False


class BingoFieldInline(admin.TabularInline):
    model = BingoField
    list_display = ("word", "position", "vote")
    readonly_fields = ("word", "position", "vote")

    def has_delete_permission(self, request, obj=None):
        return False

    def has_add_permission(self, request, obj=None):
        return False


class BingoBoardAdmin(admin.ModelAdmin):
    list_display = (bingoboard_name, "color", "created",
                    bingoboard_game_id, bingoboard_user)
    list_editable = ("color",)
    list_filter = ("game__site",)
    # cannot be changed, because its hashed.
    # if we would hash it on save, it will be hashed again on each save
    readonly_fields = ("created", "last_used", "game", "password")
    inlines = (BingoFieldInline,)


class BingoFieldAdmin(admin.ModelAdmin):
    list_display = ("word", "board", "position")
    readonly_fields = ("board", "word", "position", "vote")


class GameAdmin(admin.ModelAdmin):
    list_display = (game_id, "site", "created", "last_used")
    list_filter = ("site",)
    readonly_fields = ("created", "last_used")


admin.site.register(Word, WordAdmin)
admin.site.register(Game, GameAdmin)
admin.site.register(BingoBoard, BingoBoardAdmin)
admin.site.register(BingoField, BingoFieldAdmin)
